"""Implementaion of forcing workflows."""
import logging
import re
from typing import Union

import numpy as np
import pandas as pd
import xarray as xr
import xarray.core.resample

from .._compat import HAS_PYET

if HAS_PYET:
    import pyet

logger = logging.getLogger(__name__)


def precip(
    precip,
    da_like,
    clim=None,
    freq=None,
    reproj_method="nearest_index",
    resample_kwargs=None,
    logger=logger,
):
    """Return the lazy reprojection of precipitation to model.

      Applies the projection to the grid and resampling of time dimension to frequency.

    Parameters
    ----------
    precip: xarray.DataArray
        DataArray of precipitation forcing [mm]
    da_like: xarray.DataArray or Dataset
        DataArray of the target resolution and projection.
    clim: xarray.DataArray
        DataArray of monthly precipitation climatology. If provided this is used to
        to correct the precip downscaling.
    freq: str, Timedelta
        output frequency of time dimension
    reproj_method: str, optional
        Method for spatital reprojection of precip, by default 'nearest_index'
    resample_kwargs:
        Additional key-word arguments (e.g. label, closed) for time resampling method
    kwargs:
        Additional arguments to pass through to underlying methods.
    logger:
        The logger to use

    Returns
    -------
    p_out: xarray.DataArray (lazy)
        processed precipitation forcing
    """
    resample_kwargs = resample_kwargs or {}
    if precip.raster.dim0 != "time":
        raise ValueError(f'First precip dim should be "time", not {precip.raster.dim0}')
    # downscale precip (lazy); global min of zero
    p_out = np.fmax(precip.raster.reproject_like(da_like, method=reproj_method), 0)
    # correct precip based on high-res monthly climatology
    if clim is not None:
        # make sure first dim is month
        if clim.raster.dim0 != "month":
            clim = clim.rename({clim.raster.dim0: "month"})
        if not clim["month"].size == 12:
            raise ValueError("Precip climatology does not contain 12 months.")
        # set missings to NaN
        clim = clim.raster.mask_nodata()
        # calculate downscaling multiplication factor
        clim_coarse = clim.raster.reproject_like(
            precip, method="average"
        ).raster.reproject_like(da_like, method="average")
        clim_fine = clim.raster.reproject_like(da_like, method="average")
        p_mult = xr.where(clim_coarse > 0, clim_fine / clim_coarse, 1.0).fillna(1.0)
        # multiply with monthly multiplication factor
        p_out = p_out.groupby("time.month") * p_mult
    # resample time
    p_out.name = "precip"
    p_out.attrs.update(unit="mm")
    if freq is not None:
        resample_kwargs.update(upsampling="bfill", downsampling="sum", logger=logger)

        p_out = resample_time(p_out, freq, conserve_mass=True, **resample_kwargs)
    return p_out


# use dem_model (from staticmaps) and dem_forcing (meteo) in ini file
def temp(
    temp,
    dem_model,
    dem_forcing=None,
    lapse_correction=True,
    freq=None,
    reproj_method="nearest_index",
    lapse_rate=-0.0065,
    resample_kwargs=None,
    logger=logger,
):
    """Return lazy reprojection of temperature to model grid.

    Use lapse_rate for downscaling, and resampling of time
    dimension to frequency.

    Parameters
    ----------
    temp: xarray.DataArray
        DataArray of temperature forcing [°C]
    dem_model: xarray.DataArray
        DataArray of the target resolution and projection, contains elevation
        data
    dem_forcing: xarray.DataArray
        DataArray of elevation at forcing resolution. If provided this is used
        with `dem_model` to correct the temperature downscaling using a lapse rate
    lapse_correction : bool, optional
        If True, temperature is correctured based on lapse rate, by default True.
    freq: str, Timedelta
        output frequency of timedimension
    reproj_method: str, optional
        Method for spatital reprojection of precip, by default 'nearest_index'
    lapse_rate: float, optional
        lapse rate of temperature [C m-1] (default: -0.0065)
    resample_kwargs:
        Additional key-word arguments (e.g. label, closed) for time resampling method
    logger:
        The logger to use.

    Returns
    -------
    t_out: xarray.DataArray (lazy)
        processed temperature forcing
    """
    resample_kwargs = resample_kwargs or {}
    if temp.raster.dim0 != "time":
        raise ValueError(f'First temp dim should be "time", not {temp.raster.dim0}')
    # apply lapse rate
    if lapse_correction:
        # if dem_forcing is not provided, reproject dem_model
        dem_model = dem_model.raster.mask_nodata()
        if dem_forcing is None:
            dem_forcing = dem_model.raster.reproject_like(temp, "average")
            if np.any(np.isnan(dem_forcing)):
                logger.warning(
                    "Temperature lapse rate could be computed for some (edge) cells. "
                    "Consider providing a full coverage dem_forcing."
                )
        else:
            # assume nans in dem_forcing occur above the ocean only -> set to zero
            dem_forcing = dem_forcing.raster.mask_nodata().fillna(0)
            dem_forcing = dem_forcing.raster.reproject_like(temp, "average")
        # compute temperature at quasi MSL
        t_add_sea_level = temp_correction(dem_forcing, lapse_rate=lapse_rate)
        temp = temp - t_add_sea_level
    # downscale temperature (lazy) and add zeros with mask to mask areas outside AOI
    t_out = temp.raster.reproject_like(dem_model, method=reproj_method)
    if lapse_correction:
        # correct temperature based on high-res DEM
        # calculate downscaling addition
        t_add_elevation = temp_correction(dem_model, lapse_rate=lapse_rate)
        t_out = t_out + t_add_elevation
    # resample time
    t_out.name = "temp"
    t_out.attrs.update(unit="degree C.")
    if freq is not None:
        resample_kwargs.update(upsampling="bfill", downsampling="mean", logger=logger)
        t_out = resample_time(t_out, freq, conserve_mass=False, **resample_kwargs)
    return t_out


def press(
    press,
    dem_model,
    lapse_correction=True,
    freq=None,
    reproj_method="nearest_index",
    lapse_rate=-0.0065,
    resample_kwargs=None,
    logger=logger,
):
    """Return lazy reprojection of pressure to model grid.

    Resample time dimension to frequency.

    Parameters
    ----------
    press: xarray.DataArray
        DataArray of pressure forcing [hPa]
    dem_model: xarray.DataArray
        DataArray of the target resolution and projection, contains elevation
        data
    lapse_correction: str, optional
       If True 'dem_model` is used to correct the pressure with the `lapse_rate`.
    freq: str, Timedelta
        output frequency of timedimension
    reproj_method: str, optional
        Method for spatital reprojection of precip, by default 'nearest_index'
    lapse_rate: float, optional
        lapse rate of temperature [C m-1] (default: -0.0065)
    resample_kwargs:
        Additional key-word arguments (e.g. label, closed) for time resampling method
    logger:
        The logger to use.

    Returns
    -------
    press_out: xarray.DataArray (lazy)
        processed pressure forcing
    """
    if resample_kwargs is None:
        resample_kwargs = {}
    if press.raster.dim0 != "time":
        raise ValueError(f'First press dim should be "time", not {press.raster.dim0}')
    # downscale pressure (lazy)
    press_out = press.raster.reproject_like(dem_model, method=reproj_method)
    # correct temperature based on high-res DEM
    if lapse_correction:
        # calculate downscaling addition
        press_factor = press_correction(dem_model, lapse_rate=lapse_rate)
        press_out = press_out * press_factor
    # resample time
    press_out.name = "press"
    press_out.attrs.update(unit="hPa")
    if freq is not None:
        resample_kwargs.update(upsampling="bfill", downsampling="mean", logger=logger)
        press_out = resample_time(
            press_out, freq, conserve_mass=False, **resample_kwargs
        )
    return press_out


def wind(
    da_model: Union[xr.DataArray, xr.Dataset],
    wind: xr.DataArray = None,
    wind_u: xr.DataArray = None,
    wind_v: xr.DataArray = None,
    altitude: float = 10,
    altitude_correction: bool = False,
    freq: pd.Timedelta = None,
    reproj_method: str = "nearest_index",
    resample_kwargs: dict = None,
    logger=logger,
):
    """Return lazy reprojection of wind speed to model grid.

    Resample time dimension to frequency. Either provides wind speed directly
    or both wind_u and wind_v components.

    Parameters
    ----------
    wind: xarray.DataArray
        DataArray of wind speed forcing [m s-1]
    wind_u: xarray.DataArray
        DataArray of U component of wind speed forcing [m s-1]
    wind_v: xarray.DataArray
        DataArray of V component of wind speed forcing [m s-1]
    da_model: xarray.DataArray
        DataArray of the target resolution and projection
    altitude: float, optional
        ALtitude of wind speed data. By default 10m.
    altitude_correction: str, optional
       If True wind speed is re-calculated to wind speed at 2 meters using
       original `altitude`.
    freq: str, Timedelta
        output frequency of timedimension
    reproj_method: str, optional
        Method for spatital reprojection of precip, by default 'nearest_index'
    resample_kwargs:
        Additional key-word arguments (e.g. label, closed) for time resampling method
    logger:
        The logger to use.

    Returns
    -------
    wind_out: xarray.DataArray (lazy)
        processed wind forcing
    """
    if resample_kwargs is None:
        resample_kwargs = {}
    if wind_u is not None and wind_v is not None:
        wind = np.sqrt(np.power(wind_u, 2) + np.power(wind_v, 2))
    elif wind is None:
        raise ValueError("Either wind or wind_u and wind_v varibales must be supplied.")

    if wind.raster.dim0 != "time":
        raise ValueError(f'First wind dim should be "time", not {wind.raster.dim0}')

    # compute wind at 2 meters altitude
    if altitude_correction:
        wind = wind * (4.87 / np.log((67.8 * altitude) - 5.42))
    # downscale wind (lazy)
    wind_out = wind.raster.reproject_like(da_model, method=reproj_method)
    # resample time
    wind_out.name = "wind"
    wind_out.attrs.update(unit="m s-1")
    if freq is not None:
        resample_kwargs.update(upsampling="bfill", downsampling="mean", logger=logger)
        wind_out = resample_time(wind_out, freq, conserve_mass=False, **resample_kwargs)
    return wind_out


def pet(
    ds: xarray.Dataset,
    temp: xarray.DataArray,
    dem_model: xarray.DataArray,
    method: str = "debruin",
    press_correction: bool = False,
    wind_correction: bool = True,
    wind_altitude: float = 10,
    reproj_method: str = "nearest_index",
    freq: str = None,
    resample_kwargs: dict = None,
    logger=logger,
) -> xarray.DataArray:
    """Determine reference evapotranspiration.

    (lazy reprojection on model grid and resampling of time dimension to frequency).

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset with climate variables: pressure [hPa], global radiation [W m-2],
        TOA incident solar radiation [W m-2], wind [m s-1]

        * Required variables: {"temp", "press" or "press_msl", "kin"}
        * additional variables for debruin: {"kout"}
        * additional variables for penman-monteith_rh_simple:
            {"temp_min", "temp_max", "wind" or "wind_u"+"wind_v", "rh"}
        * additional variables for penman-monteith_tdew:
            {"temp_min", "temp_max", "wind" or "wind_u"+"wind_v", "temp_dew"}
    temp : xarray.DataArray
        DataArray with temperature on model grid resolution [°C]
    dem_model : xarray.DataArray
        DataArray of the target resolution and projection, contains elevation
        data
    method : {'debruin', 'makkink', "penman-monteith_rh_simple", "penman-monteith_tdew"}
        Potential evapotranspiration method.
        if penman-monteith is used, requires the installation of the pyet package.
    reproj_method: str, optional
        Method for spatital reprojection of precip, by default 'nearest_index'
    press_correction : bool, default False
        If True pressure is corrected, based on elevation data of `dem_model`
    wind_altitude: float, optional
        ALtitude of wind speed data. By default 10m.
    wind_correction: str, optional
        If True wind speed is re-calculated to wind speed at 2 meters using
        original `wind_altitude`.
    freq: str, Timedelta, default None
        output frequency of timedimension
    resample_kwargs:
        Additional key-word arguments (e.g. label, closed) for time resampling method
    logger:
        The logger to use.

    Returns
    -------
    pet_out : xarray.DataArray (lazy)
        reference evapotranspiration
    """
    # # resample in time
    if resample_kwargs is None:
        resample_kwargs = {}
    if temp.raster.dim0 != "time" or ds.raster.dim0 != "time":
        raise ValueError('First dimension of input variables should be "time"')
    # make sure temp and ds align both temporally and spatially
    if not np.all(temp["time"].values == ds["time"].values):
        raise ValueError("All input variables have same time index.")
    if not temp.raster.identical_grid(dem_model):
        raise ValueError("Temp variable should be on model grid.")

    # resample input to model grid
    ds = ds.raster.reproject_like(dem_model, method=reproj_method)

    # Process bands like 'pressure' and 'wind'
    if press_correction:
        ds["press"] = press(
            ds["press_msl"],
            dem_model,
            lapse_correction=press_correction,
            freq=None,  # do not change freq of press, put pet_out later
            reproj_method=reproj_method,
        )
    else:
        if "press_msl" in ds:
            ds = ds.rename({"press_msl": "press"})
        elif HAS_PYET:
            # calculate pressure from elevation [kPa]
            ds["press"] = pyet.calc_press(dem_model)
            # convert to hPa to be consistent with press function calculation:
            ds["press"] = ds["press"] * 10
        else:
            raise ModuleNotFoundError(
                "If 'press' is supplied and 'press_correction' is not used,"
                + " the pyet package must be installed."
            )

    timestep = to_timedelta(ds).total_seconds()
    if method == "debruin":
        pet_out = pet_debruin(
            temp,
            ds["press"],
            ds["kin"],
            ds["kout"],
            timestep=timestep,
        )
    elif method == "makkink":
        pet_out = pet_makkink(temp, ds["press"], ds["kin"], timestep=timestep)
    elif "penman-monteith" in method:
        logger.info("Calculating Penman-Monteith ref evaporation")
        # Add wind
        # compute wind from u and v components at 10m (for era5)
        if ("wind10_u" in ds.data_vars) & ("wind10_v" in ds.data_vars):
            ds["wind"] = wind(
                da_model=dem_model,
                wind_u=ds["wind10_u"],
                wind_v=ds["wind10_v"],
                altitude=wind_altitude,
                altitude_correction=wind_correction,
            )
        else:
            ds["wind"] = wind(
                da_model=dem_model,
                wind=ds["wind"],
                altitude=wind_altitude,
                altitude_correction=wind_correction,
            )
        if method == "penman-monteith_rh_simple":
            pet_out = pm_fao56(
                temp["temp"],
                temp["temp_max"],
                temp["temp_min"],
                ds["press"],
                ds["kin"],
                ds["wind"],
                ds["rh"],
                dem_model,
                "rh",
            )
        elif method == "penman-monteith_tdew":
            pet_out = pm_fao56(
                temp["temp"],
                temp["temp_max"],
                temp["temp_min"],
                ds["press"],
                ds["kin"],
                ds["wind"],
                ds["temp_dew"],
                dem_model,
                "temp_dew",
            )
        else:
            methods = [
                "debruin",
                "makking",
                "penman-monteith_rh_simple",
                "penman-monteith_tdew",
            ]
            raise ValueError(f"Unknown pet method, select from {methods}")

    # resample in time
    pet_out.name = "pet"
    pet_out.attrs.update(unit="mm")
    if freq is not None:
        resample_kwargs.update(upsampling="bfill", downsampling="sum", logger=logger)
        pet_out = resample_time(pet_out, freq, conserve_mass=True, **resample_kwargs)
    return pet_out


def press_correction(
    dem_model, g=9.80665, R_air=8.3144621, Mo=0.0289644, lapse_rate=-0.0065
):
    """Pressure correction based on elevation lapse_rate.

    Parameters
    ----------
    dem_model : xarray.DataArray
        DataArray with high res lat/lon axis and elevation data
    g : float, default 9.80665
        gravitational constant [m s-2]
    R_air : float, default 8.3144621
        specific gas constant for dry air [J mol-1 K-1]
    Mo : float, default 0.0289644
        molecular weight of gas [kg / mol]
    lapse_rate : float, deafult -0.0065
        lapse rate of temperature [C m-1]

    Returns
    -------
    press_fact : xarray.DataArray
        pressure correction factor
    """
    # constant
    pow = g * Mo / (R_air * lapse_rate)
    press_fact = np.power(288.15 / (288.15 + lapse_rate * dem_model), pow).fillna(1.0)
    return press_fact


def temp_correction(dem, lapse_rate=-0.0065):
    """Temperature correction based on elevation data.

    Parameters
    ----------
    dem : xarray.DataArray
        DataArray with elevation
    lapse_rate : float, default -0.0065
        lapse rate of temperature [°C m-1]

    Returns
    -------
    temp_add : xarray.DataArray
        temperature addition
    """
    temp_add = (dem * lapse_rate).fillna(0)

    return temp_add


def pet_debruin(
    temp, press, k_in, k_ext, timestep=86400, cp=1005.0, beta=20.0, Cs=110.0
):
    """Determine De Bruin (2016) reference evapotranspiration.

    Parameters
    ----------
    temp : xarray.DataArray
        DataArray with temperature [°C]
    press : xarray.DataArray
        pressure at surface [hPa]
    k_in : xarray.DataArray
        global (=short wave incoming) radiation [W m-2]
    k_ext : xarray.DataArray
        TOA incident solar radiation [W m-2]
    timestep : int, default 86400
        seconds per timestep
    cp : float, default 1005.0
        standard cp [J kg-1 K-1]
    beta : float, default 20.0
        correction constant [W m-2]
    Cs : float, default 110.0
        emperical constant [W m-2]

    Returns
    -------
    pet : xarray.DataArray
        reference evapotranspiration
    """
    # saturation and actual vapour pressure at given temperature [hPa °C-1]
    esat = 6.112 * np.exp((17.67 * temp) / (temp + 243.5))
    # slope of vapour pressure curve [hPa °C-1]
    slope = esat * (17.269 / (temp + 243.5)) * (1.0 - (temp / (temp + 243.5)))
    # compute latent heat of vapourization [J kg-1]
    lam = (2.502 * 10**6) - (2250.0 * temp)
    # psychometric constant [hPa °C-1]
    gamma = (cp * press) / (0.622 * lam)
    # compute ref. evaporation (with global radiation, therefore calling it potential)
    # in J m-2 over whole period
    ep_joule = (
        (slope / (slope + gamma))
        * (((1.0 - 0.23) * k_in) - (Cs * (k_in / (k_ext + 0.00001))))
    ) + beta
    ep_joule = xr.where(k_ext == 0.0, 0.0, ep_joule)
    pet = ((ep_joule / lam) * timestep).astype(np.float32)
    pet = xr.where(pet > 0.0, pet, 0.0)
    return pet


def pet_makkink(temp, press, k_in, timestep=86400, cp=1005.0):
    """Determnines Makkink reference evapotranspiration.

    Parameters
    ----------
    temp : xarray.DataArray
        DataArray with temperature [°C]
    press : xarray.DataArray
        DataArray with pressure [hPa]
    k_in : xarray.DataArray
        DataArray with global radiation [W m-2]
    timestep : int, default 86400
        seconds per timestep
    cp : float, default 1005.0
        standard cp [J kg-1 K-1]

    Returns
    -------
    pet : xarray.DataArray (lazy)
        reference evapotranspiration
    """
    # saturation and actual vapour pressure at given temperature [hPa °C-1]
    esat = 6.112 * np.exp((17.67 * temp) / (temp + 243.5))
    # slope of vapour pressure curve [hPa °C-1]
    slope = esat * (17.269 / (temp + 243.5)) * (1.0 - (temp / (temp + 243.5)))
    # compute latent heat of vapourization [J kg-1]
    lam = (2.502 * 10**6) - (2250.0 * temp)
    # psychometric constant [hPa °C-1]
    gamma = (cp * press) / (0.622 * lam)

    ep_joule = 0.65 * slope / (slope + gamma) * k_in
    pet = ((ep_joule / lam) * timestep).astype(np.float32)
    pet = xr.where(pet > 0.0, pet, 0.0)
    return pet


def pm_fao56(
    temp: xarray.DataArray,
    temp_max: xarray.DataArray,
    temp_min: xarray.DataArray,
    press: xarray.DataArray,
    kin: xarray.DataArray,
    wind: xarray.DataArray,
    temp_dew: xarray.DataArray,
    dem: xarray.DataArray,
    var: str = "temp_dew",
) -> xarray.DataArray:
    """Estimate daily reference evapotranspiration (ETo).

    Based on a hypothetical short grass reference surface using the
    FAO-56 Penman-Monteith equation. Actual vapor pressure is derived either
    from relative humidity or dewpoint temperature (depending on var_for_avp_name).
    Based on equation 6 in Allen et al (1998) and using the functions provided
    by the pyet package ()

    Parameters
    ----------
    temp : xarray.DataArray
        DataArray with daily temperature [°C]
    temp_max : xarray.DataArray
        DataArray with maximum daily temperature [°C]
    temp_min : xarray.DataArray
        DataArray with minimum daily temperature [°C]
    press : xarray.DataArray
        DataArray with pressure [hPa]
    kin : xarray.DataArray
        DataArray with global radiation [W m-2]
    wind : xarray.DataArray
        DataArray with wind speed at 2m above the surface [m s-1]
    temp_dew : xarray.DataArray
        DataArray with either temp_dew (dewpoint temperature at 2m above surface [°C])
        or rh (relative humidity [%]) to estimate actual vapor pressure
    dem : xarray.DataArray
        DataArray with elevation at model resolution [m]
    var : str, optional
        String with variable name used to estimate actual vapor pressure
        (chose from ["temp_dew", "rh"])

    Returns
    -------
    xarray.DataArray
        DataArray with the estimated daily reference evapotranspiration pet [mm d-1]

    Raises
    ------
    ModuleNotFoundError
        In case the pyet module is not installed

    """
    # Small check for libraries
    if not HAS_PYET:
        raise ModuleNotFoundError("Penman-Monteith FAO-56 requires the 'pyet' library")

    # Precalculated variables:
    lat = kin[kin.raster.y_dim] * (np.pi / 180)  # latitude in radians

    # Vapor pressure
    svp = pyet.calc_e0(tmean=temp)

    if var == "temp_dew":
        avp = pyet.calc_e0(tmean=temp_dew)
    elif var == "rh":
        avp = pyet.calc_e0(tmax=temp_max, tmin=temp_min, rh=temp_dew)

    # Net radiation
    dates = pyet.utils.get_index(kin)
    er = pyet.extraterrestrial_r(
        dates, lat
    )  # Extraterrestrial daily radiation [MJ m-2 d-1]
    csr = pyet.calc_rso(er, dem)  # Clear-sky solar radiation [MJ m-2 day-1]

    # Net shortwave radiation [MJ m-2 d-1]
    swr = pyet.calc_rad_short(kin * (86400 / 1e6))

    # Net longwave radiation [MJ m-2 d-1]
    lwr = pyet.calc_rad_long(
        kin * (86400 / 1e6), tmax=temp_max, tmin=temp_min, rso=csr, ea=avp
    )
    nr = swr - lwr  # daily net radiation

    # Penman Monteith FAO-56
    gamma = pyet.calc_psy(press / 10)  # psychrometric constant
    dlt = pyet.calc_vpc(
        temp
    )  # Slope of saturation vapour pressure curve at air Temperature [kPa °C-1].

    gamma1 = gamma * (1 + 0.34 * wind)

    den = dlt + gamma1
    num1 = (0.408 * dlt * (nr - 0)) / den
    num2 = (gamma * (svp - avp) * 900 * wind / (temp + 273.15)) / den
    pet = num1 + num2
    pet = pyet.utils.clip_zeros(pet, True)

    return pet.rename("pet")


def resample_time(
    da,
    freq,
    label="right",
    closed="right",
    upsampling="bfill",
    downsampling="mean",
    conserve_mass=True,
    logger=logger,
):
    """Resample data to destination frequency.

    Skip if input data already at output frequency.

    Parameters
    ----------
    da: xarray.DataArray
        Input data
    freq: str, pd.timedelta
        Output frequency.
    label: {'left', 'right'}, optional
        Side of each interval to use for labeling. By default 'right'.
    closed: {'left', 'right'}, optional
        Side of each interval to treat as closed. By default 'right'.
    upsampling, downsampling: str, optional
        Resampling method if output frequency is higher, lower (resp.) compared
        to input frequency.
    conserve_mass: bool, optional
        If True multiply output with relative change in frequency to conserve mass
    logger:
        The logger to use.

    Returns
    -------
    pet : xarray.DataArray
        Resampled data.
    """
    da_out = da
    dfreq = delta_freq(da, freq)
    if not np.isclose(dfreq, 1.0):
        resample = upsampling if dfreq < 1 else downsampling
        pre = "up" if dfreq < 1 else "down"
        logger.debug(
            f"{pre}sampling {da.name} using {resample}; conserve mass: {conserve_mass}"
        )
        if not hasattr(xr.core.resample.DataArrayResample, resample):
            raise ValueError(f"unknown resampling option {resample}")
        da_resampled = da.resample(time=freq, skipna=True, label=label, closed=closed)
        da_out = getattr(da_resampled, resample)()
        if conserve_mass:
            da_out = da_out * min(dfreq, 1)

    return da_out


def delta_freq(da_or_freq, da_or_freq1):
    """Return relative difference between dataset mean timestep and destination freq.

    <1 : upsampling
    1 : same
    >1 : downsampling.
    """
    return to_timedelta(da_or_freq1) / to_timedelta(da_or_freq)


def to_timedelta(da_or_freq):
    """Convert time dimension or frequency to timedelta."""
    if isinstance(da_or_freq, (xr.DataArray, xr.Dataset)):
        freq = da_to_timedelta(da_or_freq)
    else:
        freq = freq_to_timedelta(da_or_freq)
    return freq


def da_to_timedelta(da):
    """Convert time dimenstion in dataset to timedelta."""
    return pd.to_timedelta(np.diff(da.time).mean())


def freq_to_timedelta(freq):
    """Convert frequency to timedelta."""
    # Add '1' to freq that doesn't have any digit
    if isinstance(freq, str) and not bool(re.search(r"\d", freq)):
        freq = "1{}".format(freq)

    # Convert str to datetime.timedelta
    return pd.to_timedelta(freq)



"""Implementations relating to merging of datasets and arrays."""
import numpy as np
import rasterio
from pyproj import CRS

from .raster import full_from_transform


def merge(
    data_arrays,
    dst_crs=None,
    dst_bounds=None,
    dst_res=None,
    align=True,
    mask=None,
    merge_method="first",
    **kwargs,
):
    """Merge multiple tiles to a single DataArray.

    If mismatching grid CRS or resolution, tiles are reprojected
    to match the output DataArray grid. If no destination
    grid is defined it is based on the first DataArray in the list.

    Based on :py:meth:`rasterio.merge.merge`.

    Arguments
    ---------
    data_arrays: list of xarray.DataArray
        Tiles to merge
    dst_crs: pyproj.CRS, int
        CRS (or EPSG code) of destination grid
    dst_bounds: list of float
        Bounding box [xmin, ymin, xmax, ymax] of destination grid
    dst_res: float
        Resolution of destination grid
    align : bool, optional
        If True, align grid with dst_res
    mask: geopands.GeoDataFrame, optional
        Mask of destination area of interest.
        Used to determine dst_crs, dst_bounds and dst_res if missing.
    merge_method: {'first','last','min','max','new'}, callable
        Merge method:

        * first: reverse painting
        * last: paint valid new on top of existing
        * min: pixel-wise min of existing and new
        * max: pixel-wise max of existing and new
        * new: assert no pixel overlap
    **kwargs:
        Key-word arguments passed to
        :py:meth:`~hydromt.raster.RasterDataArray.reproject`

    Returns
    -------
    da_out: xarray.DataArray
        Merged tiles.
    """
    # define merge method
    if merge_method == "first":

        def copyto(old_data, new_data, old_nodata, new_nodata):
            dmask = np.logical_and(old_nodata, ~new_nodata)
            old_data[dmask] = new_data[dmask]

    elif merge_method == "last":

        def copyto(old_data, new_data, old_nodata, new_nodata):
            dmask = ~new_nodata
            old_data[dmask] = new_data[dmask]

    elif merge_method == "min":

        def copyto(old_data, new_data, old_nodata, new_nodata):
            dmask = np.logical_and(~old_nodata, ~new_nodata)
            old_data[dmask] = np.minimum(old_data[dmask], new_data[dmask])
            dmask = np.logical_and(old_nodata, ~new_nodata)
            old_data[dmask] = new_data[dmask]

    elif merge_method == "max":

        def copyto(old_data, new_data, old_nodata, new_nodata):
            dmask = np.logical_and(~old_nodata, ~new_nodata)
            old_data[dmask] = np.maximum(old_data[dmask], new_data[dmask])
            dmask = np.logical_and(old_nodata, ~new_nodata)
            old_data[dmask] = new_data[dmask]

    elif merge_method == "new":

        def copyto(old_data, new_data, old_nodata, new_nodata):
            data_overlap = np.logical_and(~old_nodata, ~new_nodata)
            assert not np.any(data_overlap)
            mask = ~new_nodata
            old_data[mask] = new_data[mask]

    elif callable(merge_method):
        copyto = merge_method

    # resolve arguments
    if data_arrays[0].ndim != 2:
        raise ValueError("Mosaic is only implemented for 2D DataArrays.")
    if dst_crs is None and (dst_bounds is not None or dst_res is not None):
        raise ValueError("dst_bounds and dst_res not understood without dst_crs.")
    idx0 = 0
    if mask is not None and (dst_res is None or dst_crs is None):
        # select array with largest overlap of valid cells if mask
        areas = []
        for da in data_arrays:
            da_clipped = da.raster.clip_geom(geom=mask, mask=True)
            n = da_clipped.raster.mask_nodata().notnull().sum().load().item()
            areas.append(n)
        idx0 = np.argmax(areas)
        # if single layer with overlap of valid cells: return clip
        if np.sum(np.array(areas) > 0) == 1:
            return data_arrays[idx0].raster.clip_geom(mask)
    da0 = data_arrays[idx0]  # used for default dst_crs and dst_res
    # dst CRS
    if dst_crs is None:
        dst_crs = da0.raster.crs
    else:
        dst_crs = CRS.from_user_input(dst_crs)
    # dst res
    if dst_res is None:
        dst_res = da0.raster.res
    if isinstance(dst_res, float):
        x_res, y_res = dst_res, -dst_res
    elif isinstance(dst_res, (tuple, list)) and len(dst_res) == 2:
        x_res, y_res = dst_res
    else:
        raise ValueError("dst_res not understood.")
    assert x_res > 0
    # TODO test y_res > 0
    dst_res = (x_res, -y_res)  # NOTE: y_res is multiplied with -1 in rasterio!
    # dst bounds
    if dst_bounds is None and mask is not None:
        w, s, e, n = mask.to_crs(dst_crs).buffer(2 * abs(x_res)).total_bounds
        # align with da0 grid
        w0, s0, e0, n0 = da0.raster.bounds
        w = w0 + np.round((w - w0) / abs(x_res)) * abs(x_res)
        n = n0 + np.round((n - n0) / abs(y_res)) * abs(y_res)
        e = w + int(round((e - w) / abs(x_res))) * abs(x_res)
        s = n - int(round((n - s) / abs(y_res))) * abs(y_res)
        align = False  # don't align with resolution
    elif dst_bounds is None:
        for i, da in enumerate(data_arrays):
            if i == 0:
                w, s, e, n = da.raster.transform_bounds(dst_crs)
            else:
                w1, s1, e1, n1 = da.raster.transform_bounds(dst_crs)
                w = min(w, w1)
                s = min(s, s1)
                e = max(e, e1)
                n = max(n, n1)
    else:
        w, s, e, n = dst_bounds
    # check N>S orientation
    top, bottom = (n, s) if y_res < 0 else (s, n)
    dst_bounds = w, bottom, e, top
    # dst transform
    width = int(round((e - w) / abs(x_res)))
    height = int(round((n - s) / abs(y_res)))
    transform = rasterio.transform.from_bounds(*dst_bounds, width, height)
    # align with dst_res
    if align:
        transform, width, height = rasterio.warp.aligned_target(
            transform, width, height, dst_res
        )

    # creat output array
    nodata = da0.raster.nodata
    isnan = np.isnan(nodata)
    dtype = da0.dtype
    da_out = full_from_transform(
        transform=transform,
        shape=(height, width),
        nodata=nodata,
        dtype=dtype,
        name=da0.name,
        attrs=da0.attrs,
        crs=dst_crs,
    )
    ys = da_out.raster.ycoords.values
    xs = da_out.raster.xcoords.values
    dest = da_out.values
    kwargs.update(dst_crs=dst_crs, dst_res=dst_res, align=True)
    sf_lst = []
    for _, da in enumerate(data_arrays):
        if not da.raster.aligned_grid(da_out):
            # clip with buffer
            src_bbox = da_out.raster.transform_bounds(da.raster.crs)
            da = da.raster.clip_bbox(src_bbox, buffer=10)
            if np.any([da[dim].size == 0 for dim in da.raster.dims]):
                continue  # out of bounds
            # reproject
            da = da.raster.reproject(**kwargs)
        # clip to dst_bounds
        da = da.raster.clip_bbox(dst_bounds)
        if np.any([da[dim].size == 0 for dim in da.raster.dims]):
            continue  # out of bounds
        if "source_file" in da.attrs:
            sf_lst.append(da.attrs["source_file"])
        # merge overlap
        w0, s0, e0, n0 = da.raster.bounds
        if y_res < 0:
            top = np.where(ys <= n0)[0][0] if n0 < ys[0] else 0
            bottom = np.where(ys < s0)[0][0] if s0 > ys[-1] else None
        else:
            top = np.where(ys > n0)[0][0] if n0 < ys[-1] else 0
            bottom = np.where(ys >= s0)[0][0] if s0 > ys[0] else None
        left = np.where(xs >= w0)[0][0] if w0 > xs[0] else 0
        right = np.where(xs > e0)[0][0] if e0 < xs[-1] else None
        y_slice = slice(top, bottom)
        x_slice = slice(left, right)
        region = dest[y_slice, x_slice]
        temp = np.ma.masked_equal(da.values.astype(dtype), nodata)
        if isnan:
            region_nodata = np.isnan(region)
            temp_nodata = np.isnan(temp)
        else:
            region_nodata = region == nodata
            temp_nodata = temp.mask
        copyto(region, temp, region_nodata, temp_nodata)

    # set attrs
    da_out.attrs.update(source_file="; ".join(sf_lst))
    da_out.raster.set_crs(dst_crs)
    return da_out.raster.reset_spatial_dims_attrs()



"""Implementation for mesh based workflows."""
import logging
from typing import Dict, List, Optional, Union

import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr
import xugrid as xu
from pyproj import CRS
from shapely.geometry import box
from xugrid.ugrid import conventions

from .. import gis_utils
from ..raster import GEO_MAP_COORD
from .basin_mask import parse_region

logger = logging.getLogger(__name__)

__all__ = [
    "create_mesh2d",
    "mesh2d_from_rasterdataset",
    "mesh2d_from_raster_reclass",
    "rename_mesh",
]


def create_mesh2d(
    region: Dict,
    res: Optional[float] = None,
    crs: Optional[int] = None,
    align: bool = True,
    logger=logger,
) -> xu.UgridDataset:
    """
    Create an 2D unstructured mesh or reads an existing 2D mesh.

    Grids are read according to UGRID conventions. An 2D unstructured mesh
    will be created as 2D rectangular grid from a geometry (geom_fn) or bbox.
    If an existing 2D mesh is given, then no new mesh will be generated but an extent
    can be extracted using the `bounds` argument of region.

    Note Only existing meshed with only 2D grid can be read.

    Adds/Updates model layers:

    * **grid_name** mesh topology: add grid_name 2D topology to mesh object

    Parameters
    ----------
    region : dict
        Dictionary describing region of interest, bounds can be provided
        for type 'mesh'. In case of 'mesh', if the file includes several
        grids, the specific 2D grid can be selected using the 'grid_name' argument.
        CRS for 'bbox' and 'bounds' should be 4326; e.g.:

        * {'bbox': [xmin, ymin, xmax, ymax]}

        * {'geom': 'path/to/polygon_geometry'}

        * {'mesh': 'path/to/2dmesh_file'}

        * {'mesh': 'path/to/2dmesh_file', 'grid_name': 'mesh2d', 'bounds': [xmin, ymin, xmax, ymax]}
    res: float
        Resolution used to generate 2D mesh [unit of the CRS], required if region
        is not based on 'mesh'.
    crs : EPSG code, int, optional
        Optional EPSG code of the model or "utm" to let hydromt find the closest
        projected CRS. If None using the one from region, and else 4326.
    align : bool, optional
        Align the mesh to the resolution, by default True.

    Returns
    -------
    mesh2d : xu.UgridDataset
        Generated mesh2d.
    """  # noqa: E501
    kind, region = parse_region(region, logger=logger)
    if kind != "mesh":
        if not isinstance(res, (int, float)):
            raise ValueError("res argument required for kind 'bbox', 'geom'")
        if kind == "bbox":
            bbox = region["bbox"]
            geom = gpd.GeoDataFrame(geometry=[box(*bbox)], crs=4326)
        elif kind == "geom":
            geom = region["geom"]
            if geom.crs is None:
                raise ValueError('Model region "geom" has no CRS')
        else:
            raise ValueError(
                "Region for mesh must be of kind [bbox, geom, mesh], "
                f"kind {kind} not understood."
            )
        # Parse crs and reproject geom if needed
        if crs is not None:
            crs = gis_utils.parse_crs(crs, bbox=geom.total_bounds)
            geom = geom.to_crs(crs)
        # Generate grid based on res for region bbox
        xmin, ymin, xmax, ymax = geom.total_bounds
        if align:
            xmin = round(xmin / res) * res
            ymin = round(ymin / res) * res
            xmax = round(xmax / res) * res
            ymax = round(ymax / res) * res
        # note we flood the number of faces within bounds
        ncol = round((xmax - xmin) / res)  # int((xmax - xmin) // res)
        nrow = round((ymax - ymin) / res)  # int((ymax - ymin) // res)
        dx, dy = res, -res
        faces = []
        for i in range(nrow):
            top = ymax + i * dy
            bottom = ymax + (i + 1) * dy
            for j in range(ncol):
                left = xmin + j * dx
                right = xmin + (j + 1) * dx
                faces.append(box(left, bottom, right, top))
        grid = gpd.GeoDataFrame(geometry=faces, crs=geom.crs)
        # If needed clip to geom
        if kind != "bbox":
            grid = grid.loc[
                gpd.sjoin(
                    grid, geom, how="left", predicate="intersects"
                ).index_right.notna()
            ].reset_index()
        # Create mesh from grid
        grid.index.name = "mesh2d_nFaces"
        mesh2d = xu.UgridDataset.from_geodataframe(grid)
        mesh2d = mesh2d.ugrid.assign_face_coords()
        mesh2d.ugrid.grid.set_crs(grid.crs)

    else:
        # Read existing mesh
        uds = region["mesh"]

        # Select grid and/or check single grid
        grids = dict()
        for grid in uds.grids:
            grids[grid.name] = grid
        # Select specific topology if given
        if "grid_name" in region:
            if region["grid_name"] not in grids:
                raise ValueError(
                    f"Mesh file does not contain grid {region['grid_name']}."
                )
            grid = grids[region["grid_name"]]
        elif len(grids) == 1:
            grid = list(grids.values())[0]
        else:
            raise ValueError(
                "Mesh file contains several grids. "
                "Use grid_name argument of region to select a single one."
            )
        # Chek if 2d mesh file else throw error
        if grid.topology_dimension != 2:
            raise ValueError("Grid in mesh file for create_mesh2d is not 2D.")
        # Continues with a 2D grid
        mesh2d = xu.UgridDataset(grid.to_dataset())

        # Check crs and reproject to model crs
        grid_crs = grid.crs
        if crs is not None:
            bbox = None
            if grid_crs is not None:
                if grid_crs.to_epsg() == 4326:
                    bbox = mesh2d.ugrid.grid.bounds
            crs = gis_utils.parse_crs(crs, bbox=bbox)
        else:
            crs = CRS.from_user_input(4326)
        if grid_crs is not None:  # parse crs
            mesh2d.ugrid.grid.set_crs(grid_crs)
        else:
            # Assume model crs
            logger.warning(
                "Mesh data from mesh file doesn't have a CRS."
                f" Assuming crs option {crs}"
            )
            mesh2d.ugrid.grid.set_crs(crs)
        mesh2d = mesh2d.drop_vars(GEO_MAP_COORD, errors="ignore")

        # If bounds are provided in region, extract mesh for bounds
        if "bounds" in region:
            bounds = gpd.GeoDataFrame(geometry=[box(*region["bounds"])], crs=4326)
            xmin, ymin, xmax, ymax = bounds.to_crs(mesh2d.ugrid.grid.crs).total_bounds
            subset = mesh2d.ugrid.sel(y=slice(ymin, ymax), x=slice(xmin, xmax))
            # Check if still cells after clipping
            err = (
                "MeshDataset: No data within model region."
                "Check that bounds were given in the correct crs 4326"
            )
            subset = subset.ugrid.assign_node_coords()
            if subset.ugrid.grid.node_x.size == 0 or subset.ugrid.grid.node_y.size == 0:
                raise IndexError(err)
            mesh2d = subset

    return mesh2d


def mesh2d_from_rasterdataset(
    ds: Union[xr.DataArray, xr.Dataset],
    mesh2d: Union[xu.UgridDataArray, xu.Ugrid2d],
    variables: Optional[List] = None,
    fill_method: Optional[str] = None,
    resampling_method: Optional[str] = "centroid",
    rename: Optional[Dict] = None,
    logger: logging.Logger = logger,
) -> xu.UgridDataset:
    """
    Resamples data in ds to mesh2d.

    Raster data is interpolated to the mesh using the ``resampling_method``.

    Parameters
    ----------
    ds: xr.DataArray, xr.Dataset
        Raster xarray data object.
    mesh2d: xu.UgridDataArray, xu.Ugrid2d
        Mesh2d grid to resample to.
    variables: list, optional
        List of variables to resample. By default all variables in ds.
    fill_method : str, optional
        If specified, fills no data values using fill_nodata method.
        Available methods are {'linear', 'nearest', 'cubic', 'rio_idw'}.
    resampling_method: str, optional
        Method to sample from raster data to mesh. By default mean. Options include
        {"centroid", "barycentric", "mean", "harmonic_mean", "geometric_mean", "sum",
        "minimum", "maximum", "mode", "median", "max_overlap"}. If centroid, will use
        :py:meth:`xugrid.CentroidLocatorRegridder` method. If barycentric, will use
        :py:meth:`xugrid.BarycentricInterpolator` method. If any other, will use
        :py:meth:`xugrid.OverlapRegridder` method.
        Can provide a list corresponding to ``variables``.
    rename: dict, optional
        Dictionary to rename variable names in ds
        {'name_ds': 'name_in_uds_out'}. By default empty.

    Returns
    -------
    uds_out: xu.UgridDataset
        Resampled data on mesh2d.
    """
    rename = rename or {}
    if variables is not None:
        ds = ds[variables]
    if isinstance(ds, xr.DataArray):
        ds = ds.to_dataset()

    if fill_method is not None:
        ds = ds.raster.interpolate_na(method=fill_method)

    # check resampling method
    resampling_method = np.atleast_1d(resampling_method)
    if len(resampling_method) == 1:
        resampling_method = np.repeat(resampling_method, len(ds.data_vars))
    # one reproject method per variable
    elif len(resampling_method) != len(variables):
        raise ValueError(
            f"resampling_method should have length 1 or {len(ds.data_vars)}"
        )

    # Prepare regridder
    regridder = dict()
    # Get one variable name in ds to simplify to da
    var = [v for v in ds.data_vars][0]
    uda = xu.UgridDataArray.from_structured(
        ds[var].rename({ds.raster.x_dim: "x", ds.raster.y_dim: "y"})
    )
    uda.ugrid.set_crs(ds.raster.crs)

    # Need to reproject before calling regridder
    if hasattr(mesh2d, "ugrid"):
        mesh2d_crs = mesh2d.ugrid.grid.crs
    else:  # ugrid2d
        mesh2d_crs = mesh2d.crs
    if mesh2d_crs != uda.ugrid.grid.crs:
        uda = uda.ugrid.to_crs(mesh2d_crs)

    for method in np.unique(resampling_method):
        logger.info(f"Preparing regridder for {method} method")
        if method == "centroid":
            regridder[method] = xu.CentroidLocatorRegridder(uda, mesh2d)
        elif method == "barycentric":
            regridder[method] = xu.BarycentricInterpolator(uda, mesh2d)
        else:
            regridder[method] = xu.OverlapRegridder(uda, mesh2d, method=method)

    # Convert ds to xugrid
    for i, var in enumerate(ds.data_vars):
        logger.info(f"Resampling {var} to mesh2d using {resampling_method[i]} method")
        uda = xu.UgridDataArray.from_structured(
            ds[var].rename({ds.raster.x_dim: "x", ds.raster.y_dim: "y"})
        )
        uda.ugrid.set_crs(ds.raster.crs)
        # Reproject
        if mesh2d_crs != uda.ugrid.grid.crs:
            uda = uda.ugrid.to_crs(mesh2d_crs)
        # Interpolate
        method = resampling_method[i]
        # Interpolate
        uda_out = regridder[method].regrid(uda)
        # Add to uds_out
        if i == 0:
            uds_out = uda_out.to_dataset()
        else:
            uds_out[var] = uda_out

    # Rename variables
    if rename is not None:
        uds_out = uds_out.rename(rename)

    return uds_out


def mesh2d_from_raster_reclass(
    da: xr.DataArray,
    df_vars: pd.DataFrame,
    mesh2d: Union[xu.UgridDataArray, xu.Ugrid2d],
    reclass_variables: list,
    fill_method: Optional[str] = None,
    resampling_method: Optional[Union[str, list]] = "centroid",
    rename: Optional[Dict] = None,
    logger: logging.Logger = logger,
) -> List[str]:
    """Resample data to ``mesh2d`` grid by reclassifying the data in ``da`` based on ``df_vars``.

    The reclassified raster data
    are subsequently interpolated to the mesh using `resampling_method`.

    Parameters
    ----------
    da : xr.DataArray
        Raster xarray DataArray object.
    df_vars : pd.DataFrame
        Tabular pandas dataframe object for the reclassification table of `da`.
    mesh2d: xu.UgridDataArray, xu.Ugrid2d
        Mesh2d grid to resample to.
    reclass_variables : list
        List of reclass_variables from the df_vars table to add to the
        mesh. The index column should match values in da.
    fill_method : str, optional
        If specified, fills nodata values in `da` using the `fill_method`
        method before reclassifying. Available methods are
        {'linear', 'nearest', 'cubic', 'rio_idw'}.
    resampling_method: str, list, optional
        Method to sample from raster data to mesh. By default mean. Options include
        {"centroid", "barycentric", "mean", "harmonic_mean", "geometric_mean", "sum",
        "minimum", "maximum", "mode", "median", "max_overlap"}. If centroid, will use
        :py:meth:`xugrid.CentroidLocatorRegridder` method. If barycentric, will use
        :py:meth:`xugrid.BarycentricInterpolator` method. If any other, will use
        :py:meth:`xugrid.OverlapRegridder` method.
        Can provide a list corresponding to ``reclass_variables``.
    rename : dict, optional
        Dictionary to rename variable names in `reclass_variables` before adding
        them to the mesh. The dictionary should have the form
        {'name_in_reclass_table': 'name_in_uds_out'}. By default, an empty dictionary.

    Returns
    -------
    uds_out : xu.UgridDataset
        Resampled data on mesh2d.

    See Also
    --------
    mesh2d_from_rasterdataset
    """  # noqa: E501
    rename = rename or {}

    if fill_method is not None:
        da = da.raster.interpolate_na(method=fill_method)

    # Mapping function
    ds_vars = da.raster.reclassify(reclass_table=df_vars, method="exact")

    uds_out = mesh2d_from_rasterdataset(
        ds_vars,
        mesh2d,
        variables=reclass_variables,
        fill_method=None,
        resampling_method=resampling_method,
        rename=rename,
        logger=logger,
    )

    return uds_out


def rename_mesh(mesh: Union[xu.UgridDataArray, xu.UgridDataset], name: str):
    """
    Rename all grid variables in mesh according to UGRID conventions.

    Note: adapted from xugrid.ugrid.grid.rename to also work on
    UgridDataset and UgridDataArray
    """
    # Get the old and the new names. Their keys are the same.
    old_attrs = mesh.ugrid.grid._attrs
    new_attrs = conventions.default_topology_attrs(
        name, mesh.ugrid.grid.topology_dimension
    )

    # The attrs will have some roles joined together, e.g. node_coordinates
    # will contain x and y as "mesh2d_node_x mesh2d_node_y".
    name_dict = {mesh.ugrid.grid.name: name}
    skip = ("cf_role", "long_name", "topology_dimension")
    for key, value in old_attrs.items():
        if key in new_attrs and key not in skip:
            split_new = new_attrs[key].split()
            split_old = value.split()
            if len(split_new) != len(split_old):
                raise ValueError(
                    f"Number of entries does not match on {key}: "
                    f"{split_new} versus {split_old}"
                )
            for name_key, name_value in zip(split_old, split_new):
                name_dict[name_key] = name_value

    new = mesh.copy()
    new.ugrid.grid.name = name
    new.ugrid.grid._attrs = new_attrs
    new.ugrid.grid._indexes = {
        k: name_dict[v] for k, v in new.ugrid.grid._indexes.items()
    }

    to_rename = tuple(new.data_vars) + tuple(new.coords) + tuple(new.dims)
    new = new.rename({k: v for k, v in name_dict.items() if k in to_rename})

    return new



"""Utility functions for data adapters."""
import os
from logging import Logger
from os.path import isdir, join
from typing import Optional

import numpy as np
import pandas as pd
import xarray as xr

from hydromt.typing import Data, StrPath, Variables


def netcdf_writer(
    obj: Data,
    data_root: StrPath,
    data_name: str,
    variables: Optional[Variables] = None,
    encoder: str = "zlib",
) -> str:
    """Utiliy function for writing a xarray dataset/data array to a netcdf file.

    Parameters
    ----------
    obj : xr.Dataset | xr.DataArray
        Dataset.
    data_root : str | Path
        root to write the data to.
    data_name : str
        filename to write to.
    variables : Optional[List[str]]
        list of dataset variables to write, by default None

    Returns
    -------
    fn_out: str
        Absolute path to output file
    """
    dvars = [obj.name] if isinstance(obj, xr.DataArray) else obj.data_vars
    if variables is None:
        encoding = {k: {encoder: True} for k in dvars}
        fn_out = join(data_root, f"{data_name}.nc")
        obj.to_netcdf(fn_out, encoding=encoding)
    else:  # save per variable
        if not isdir(join(data_root, data_name)):
            os.makedirs(join(data_root, data_name))
        for var in dvars:
            fn_out = join(data_root, data_name, f"{var}.nc")
            obj[var].to_netcdf(fn_out, encoding={var: {encoder: True}})
        fn_out = join(data_root, data_name, "{variable}.nc")
    return fn_out


def zarr_writer(
    obj: Data,
    data_root: StrPath,
    data_name: str,
    **kwargs,
) -> str:
    """Utiliy function for writing a xarray dataset/data array to a netcdf file.

    Parameters
    ----------
    obj : xr.Dataset | xr.DataArray
        Dataset.
    data_root : str | Path
        root to write the data to.
    data_name : str
        filename to write to.

    Returns
    -------
    fn_out: str
        Absolute path to output file
    """
    fn_out = join(data_root, f"{data_name}.zarr")
    if isinstance(obj, xr.DataArray):
        obj = obj.to_dataset()
    obj.to_zarr(fn_out, **kwargs)
    return fn_out


def shift_dataset_time(
    dt: int, ds: Data, logger: Logger, time_unit: str = "s"
) -> xr.Dataset:
    """Shifts time of a xarray dataset.

    Parameters
    ----------
    dt : int
        time delta to shift the time of the dataset
    ds : xr.Dataset
        xarray dataset
    logger : logging.Logger
        logger

    Returns
    -------
    xr.Dataset
        time shifted dataset
    """
    if (
        dt != 0
        and "time" in ds.dims
        and ds["time"].size > 1
        and np.issubdtype(ds["time"].dtype, np.datetime64)
    ):
        logger.debug(f"Shifting time labels with {dt} {time_unit}.")
        ds["time"] = ds["time"] + pd.to_timedelta(dt, unit=time_unit)
    elif dt != 0:
        logger.warning("Time shift not applied, time dimension not found.")
    return ds
